from __future__ import annotations

import click


@click.group(
    # cls=DYMGroup,
    epilog=click.style("Commands to manage websites\n", fg="magenta"),
)
@click.version_option()
def cli() -> None:
    pass


@cli.command()
def update() -> None:
    """Update this package"""
    import subprocess
    import sys
    from shutil import which

    from .config import REPO

    uv = which("uv")
    if uv:
        ret = subprocess.call([uv, "pip", "install", "-U", REPO])
    else:
        ret = subprocess.call([sys.executable, "-m", "pip", "install", "-U", REPO])
    if ret:
        click.secho(f"can't install {REPO}", fg="red")
        raise click.Abort()


@cli.command()
def repo() -> None:
    """show git repository"""

    from .config import REPO

    click.echo(REPO)


@cli.command()
def config_show() -> None:
    """Show configuration"""
    from dataclasses import fields
    from .config import get_config, Config

    config = get_config()

    n = max(len(f.name) for f in fields(Config))

    for f in fields(Config):
        k = f.name
        v = getattr(config, f.name)
        click.echo(f"{k:<{n}}: {v}")


@cli.command()
@click.option("-a", "--append", is_flag=True, help="append to file")
@click.argument("filename")
def config_dump(filename: str, append: bool) -> None:
    """Dump configuration"""
    from .utils import require_mod
    from .config import dump_to_file

    require_mod("toml")

    if not dump_to_file(filename, append):
        click.secho("can't dump configuration!", fg="red", err=True)
        raise click.Abort()


# @cli.command()
# @click.option("-p", "--with-python", is_flag=True)
# @click.option("-c", "--compile", "use_pip_compile", is_flag=True)
# @click.argument("project_dir", required=False)
def poetry_to_reqs(
    project_dir: str,
    with_python: bool,
    use_pip_compile: bool = True,
) -> None:
    """Generate a requirements.txt file from pyproject.toml [**may require toml**]"""
    import os
    import subprocess
    from contextlib import suppress
    from .utils import toml_load

    pyproject = "pyproject.toml"
    if project_dir:
        pyproject = os.path.join(project_dir, pyproject)
    if not os.path.isfile(pyproject):
        raise click.BadArgumentUsage("no pyproject.toml file!")

    def fix(req: str) -> str:
        if req.startswith("^"):
            return f">={req[1:]}"
        return req

    reqs = "\n".join(
        f"{k}{fix(v)}"
        for k, v in sorted(
            toml_load(pyproject)["tool"]["poetry"]["dependencies"].items(),
        )
        if with_python or k != "python" and isinstance(v, str)
    )
    if use_pip_compile:
        try:
            with open("requirements.in", "w", encoding="utf-8") as fp:
                click.echo(reqs, file=fp)
            subprocess.check_call(["pip-compile"])
        finally:
            with suppress(OSError):
                os.remove("requirements.in")
    else:
        click.echo(reqs)



from __future__ import annotations

import os
import re
import subprocess
from os.path import isdir
from os.path import isfile
from os.path import join
from os.path import split
from pathlib import Path
from typing import Any
from typing import Callable
from typing import IO
from typing import TextIO
from typing import TYPE_CHECKING

import click

from ..core import get_app_entrypoint
from ..core import get_static_folders_for_app
from ..core import StaticFolder
from ..core import topath
from ..templating import get_template
from ..templating import undefined_error
from ..utils import rmfiles
from ..utils import which
from .cli import config
from .utils import asgi_option
from .utils import check_app_dir
from .utils import CHECKTYPE
from .utils import config_options
from .utils import CONVERTER
from .utils import find_favicon
from .utils import fix_params
from .utils import footprint_config
from .utils import get_default_venv
from .utils import get_known
from .utils import has_error_page
from .utils import make_args
from .utils import template_option
from .utils import to_check_func
from .utils import to_output
from .utils import url_match

if TYPE_CHECKING:
    from jinja2 import Template


def run_app(
    application_dir: str,
    gunicorn: str,
    bind: str | None = None,
    pidfile: str | None = None,
    app: str = "app.app",
) -> None:
    if pidfile is None:
        pidfile = "/tmp/gunicorn.pid"

    bind = bind if bind else "unix:app.sock"

    cmd = [
        gunicorn,
        "--pid",
        "pidfile",
        "--access-logfile=-",
        "--error-logfile=-",
        "--bind",
        bind,
        app,
    ]

    click.secho(
        f"starting gunicorn in {topath(application_dir)}",
        fg="green",
        bold=True,
    )

    click.secho(" ".join(cmd), fg="green")
    subprocess.run(cmd, cwd=application_dir, env=os.environ, check=True)


def nginx_install(nginxfile: str) -> str | None:
    import filecmp
    from ..config import get_config

    Config = get_config()

    conf = split(nginxfile)[-1]
    # Ubuntu, RHEL8
    for targetd in Config.nginx_dirs:
        if isdir(targetd):
            break
    else:
        raise RuntimeError("can't find nginx configuration directory")
    sudo = which("sudo")
    systemctl = which("systemctl")

    def sudocmd(*args: str, check: bool = True) -> subprocess.CompletedProcess[bytes]:
        return subprocess.run([sudo] + list(args), check=check)

    def systemctlcmd(*args: str, check: bool = True) -> int:
        return subprocess.run([sudo, systemctl] + list(args), check=check).returncode

    exists = isfile(f"{targetd}/{conf}")
    if not exists or not filecmp.cmp(f"{targetd}/{conf}", nginxfile):
        if exists:
            click.secho(f"warning: overwriting old {conf}", fg="yellow")

        sudocmd("cp", nginxfile, f"{targetd}/")

        if sudocmd("nginx", "-t", check=False).returncode != 0:
            sudocmd("rm", f"{targetd}/{conf}", check=True)
            click.secho("nginx configuration faulty", fg="red", err=True)
            return None

        systemctlcmd("restart", "nginx")
    else:
        click.secho(f"nginx file {conf} unchanged", fg="green")
    return conf


def nginx_uninstall(nginxfile: str) -> None:
    from ..config import get_config

    Config = get_config()

    nginxfile = split(nginxfile)[-1]
    if "." not in nginxfile:
        nginxfile += ".conf"
    sudo = which("sudo")
    systemctl = which("systemctl")

    def sudocmd(*args: str, check: bool = True) -> subprocess.CompletedProcess[bytes]:
        return subprocess.run([sudo] + list(args), check=check)

    def systemctlcmd(*args: str, check: bool = True) -> int:
        return subprocess.run([sudo, systemctl] + list(args), check=check).returncode

    for d in Config.nginx_dirs:
        fname = join(d, nginxfile)
        if isfile(fname):
            sudocmd("rm", fname)
            systemctlcmd("restart", "nginx.service")
            return

    click.secho(f"no nginx file {nginxfile}", fg="yellow", err=True)


NGINX_ARGS = {
    "server_name": "name of website",
    "application_dir": "locations of repo",
    "app": "entrypoint to app",
    "appname": "application name [default: directory name]",
    "root": "static files root directory",
    "root_prefix": "location prefix to use (only used if root is defined)",
    "prefix": "url prefix for application [default: /]",
    "expires": "expires header for static files [default: off] e.g. 30d",
    "listen": "listen on port [default: 80]",
    "host": "proxy to a port [default: use unix socket]",
    "port": "TCP/IP port to use",
    "root_location_match": "regex for matching static directory files",
    "access_log": "'on' or 'off'. log static asset requests [default:off]",
    "extra": "extra (legal) nginx commands for proxy",
    "ssl": "create an secure server configuration [see nginx-ssl]",
    "log_format": "specify the log_format",
    "authentication": "authentication file",
}

NGINX_HELP = f"""
Generate a nginx conf file for website.

Use footprint config nginx website ... etc.
with the following arguments:

\b
{make_args(NGINX_ARGS)}
\b
example:
\b
footprint config nginx mcms.plantenergy.edu.au access-log=on
"""


def appname_func(params: dict[str, Any]) -> str:
    app = str(params["application_dir"])
    return str(split(app)[-1])


def nginx(  # noqa: C901
    application_dir: str,
    server_name: str,
    args: list[str] | None = None,
    *,
    template_name: str | None = None,
    help_args: dict[str, str] | None = None,
    check: bool = True,
    output: str | TextIO | None = None,
    extra_params: dict[str, Any] | None = None,
    checks: list[tuple[str, CHECKTYPE]] | None = None,
    ignore_unknowns: bool = False,
    default_values: list[tuple[str, CONVERTER]] | None = None,
    convert: dict[str, Callable[[Any], Any]] | None = None,
    ssl: bool = False,
    asgi: bool = False,
) -> str:
    """Generate an nginx configuration for application"""
    from jinja2 import UndefinedError

    if args is None:
        args = []

    if application_dir is None:
        raise click.BadParameter("no application directory")

    if help_args is None:
        help_args = NGINX_ARGS

    if convert is None:
        convert = {"root": topath}
    else:
        convert = {"root": topath, **convert}

    application_dir = topath(application_dir)
    template = get_template(template_name or "nginx.conf", application_dir)

    known = get_known(help_args) | {"staticdirs", "favicon", "error_page"}
    # directory to match with / for say /favicon.ico
    root_location_match = None
    params: dict[str, Any] = {}
    try:
        # arguments from .flaskenv
        params = {
            k: v for k, v in footprint_config(application_dir).items() if k in known
        }
        params.update(fix_params(args, convert))
        if extra_params:
            params.update(extra_params)

        prefix = params.get("prefix", "")
        if "root" in params:
            root = topath(join(application_dir, str(params["root"])))
            params["root"] = root
            rp = params.get("root_prefix", None)
            staticdirs = [StaticFolder(rp if rp is not None else prefix, root, False)]
        else:
            staticdirs = []
        # if the params have an app value use that as the entrypoint
        entrypoint = params.get("app", None)
        if entrypoint is None:
            entrypoint = get_app_entrypoint(application_dir, asgi=asgi)
        if entrypoint != "@none":
            staticdirs.extend(
                get_static_folders_for_app(
                    application_dir,
                    entrypoint=entrypoint,
                    prefix=prefix,
                ),
            )

        error_page = has_error_page(staticdirs)  # actually 404.html
        if error_page:
            params["error_page"] = error_page
        params["staticdirs"] = staticdirs
        for s in staticdirs:
            if not s.url:  # top level?
                root_location_match = url_match(s.folder)
        # need a root directory for server
        if "root" not in params and not staticdirs:
            params["root"] = topath(application_dir)
            # raise click.BadParameter("no root directory found", param_hint="params")
        # add any defaults
        defaults: list[tuple[str, CONVERTER]] = [
            ("application_dir", lambda _: application_dir),
            ("appname", appname_func),
            ("root", lambda _: staticdirs[0].folder),
            ("server_name", lambda _: server_name),
            ("ssl", lambda _: ssl),
        ]

        if default_values:
            defaults.extend(default_values)

        for key, default_func in defaults:
            if key not in params:
                v = default_func(params)
                if v is not None:
                    params[key] = v

        def isint(s: str | int) -> bool:
            return isinstance(s, int) or s.isdigit()

        if "host" in params:
            h = params["host"]
            if isint(h):
                params["host"] = "127.0.0.1"
                params["port"] = h
            else:
                if ":" in h:
                    s, h = h.rsplit(":", maxsplit=1)
                    params["host"] = s
                    params["port"] = h

        if "port" not in params:
            params["port"] = 8000

        if root_location_match is not None and "root_location_match" not in params:
            params["root_location_match"] = root_location_match
        if "favicon" not in params and not root_location_match:
            d = find_favicon(application_dir)
            if d:
                params["favicon"] = topath(join(application_dir, d))

        if "favicon" in params:
            params["favicon"] = topath(params["favicon"])

        if check:
            msg = check_app_dir(application_dir)
            if msg:
                raise click.BadParameter(msg, param_hint="application_dir")

            if not ignore_unknowns:
                extra = set(params) - known
                if extra:
                    raise click.BadParameter(
                        f"unknown arguments {extra}",
                        param_hint="params",
                    )
            failed: list[str] = []
            checks = (checks or []) + [
                to_check_func("root", isdir, '"{root}" is not a directory'),
                to_check_func("favicon", isdir, '"{favicon}" is not a directory'),
            ]
            for key, func in checks:
                if key in params:
                    v = params[key]
                    msg = func(key, v)
                    if msg is not None:
                        click.secho(
                            msg,
                            fg="yellow",
                            bold=True,
                            err=True,
                        )
                        failed.append(key)
                if failed:
                    raise click.Abort()

        res = template.render(**params)  # pylint: disable=no-member
        to_output(res, output)
        return res
    except UndefinedError as e:
        undefined_error(e, template, params)
        raise click.Abort()


# pylint: disable=too-many-locals too-many-branches
@config.command(name="nginx", help=NGINX_HELP)  # noqa: C901
@template_option
@config_options
@click.option("--ssl", is_flag=True, help="make it secure")
@click.option(
    "--no-static",
    is_flag=True,
    help="Don't try to introspect static files. (Useful for non-flask websites)",
)
@asgi_option
@click.option(
    "-d",
    "--app-dir",
    "application_dir",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    help="""location of repo or current directory""",
)
@click.argument("server_name")
@click.argument("params", nargs=-1)
def nginx_cmd(
    application_dir: str | None,
    server_name: str,
    template: str | None,
    asgi: bool,
    params: list[str],
    no_check: bool,
    output: str | None,
    ssl: bool = False,
    no_static: bool = False,
) -> None:
    """Generate nginx config file.

    PARAMS are key=value arguments for the template.
    """
    # pylint: disable=line-too-long
    # see https://www.digitalocean.com/community/tutorials/how-to-serve-flask-applications-with-gunicorn-and-nginx-on-ubuntu-20-04
    # place this in /etc/systemd/system/
    params = list(params)
    if no_static:
        params.append("app=@none")

    nginx(
        application_dir or ".",
        server_name,
        params,
        template_name=template,
        check=not no_check,
        output=output,
        ssl=ssl,
        asgi=asgi,
    )


@config.command(name="nginx-run-app")
@click.option("-p", "--port", default=5000, help="port to listen", show_default=True)
@click.option(
    "-x",
    "--no-start",
    "no_start_app",
    is_flag=True,
    help="don't start the website in background",
    show_default=True,
)
@click.option(
    "--entrypoint",
    help="web application entrypoint",
)
@asgi_option
@click.option("--browse", is_flag=True, help="open web application in browser")
@click.option(
    "-d",
    "--app-dir",
    "application_dir",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    help="""location of repo or current directory""",
)
def nginx_run_app_cmd(
    application_dir: str | None,
    port: int,
    entrypoint: str | None,
    asgi: bool,
    no_start_app: bool = False,
    browse: bool = False,
) -> None:
    """Run nginx as a non daemon process with web app in background."""
    import signal
    import uuid
    from threading import Thread

    from tempfile import gettempdir

    from ..utils import Runner, browser

    nginx_exe = which("nginx")

    if application_dir is None:
        application_dir = "."

    application_dir = topath(application_dir)
    tmplt = get_template("nginx-test.conf", application_dir)
    res = tmplt.render(  # pylint: disable=no-member
        application_dir=application_dir,
        port=port,
    )
    tmpfile = Path(gettempdir()) / f"nginx-{uuid.uuid4()}.conf"
    pidfile = str(tmpfile) + ".pid"

    app = entrypoint or get_app_entrypoint(application_dir, asgi=asgi)

    procs: list[Runner] = []
    url = f"http://127.0.0.1:{port}"
    click.secho(f"listening on {url}", fg="green", bold=True)
    if not no_start_app:
        venv = get_default_venv()
        if os.path.isdir(venv):
            gunicorn = os.path.join(venv, "bin", "gunicorn")
        else:
            gunicorn = which("gunicorn")

        bgapp = Runner(
            app,
            [gunicorn, "--pid", pidfile, "--bind", "unix:app.sock", app],
            directory=application_dir,
        )
        procs.append(bgapp)
    else:
        click.secho(
            f"expecting app: cd {application_dir} && gunicorn --bind unix:app.sock {app}",
            fg="magenta",
            bold=True,
        )
    try:
        with tmpfile.open("wt", encoding="utf-8") as fp:
            fp.write(res)
        threads = [b.start() for b in procs]
        b: Thread | None = None
        if browse:
            b = browser(url=url)
        try:
            subprocess.check_call([nginx_exe, "-c", str(tmpfile)])
        finally:
            if not no_start_app:
                with open(pidfile, encoding="utf-8") as fp:
                    pid = int(fp.read().strip())
                    os.kill(pid, signal.SIGINT)

            for thrd in threads:
                thrd.wait()
            if b:
                b.join()
    finally:
        tmpfile.unlink(missing_ok=True)
        rmfiles([pidfile])
        os.system("stty sane")


@config.command(name="nginx-run")
@click.option(
    "-p",
    "--port",
    default=5000,
    help="port for nginx to listen",
)
@click.option(
    "--entrypoint",
    help="web application entrypoint",
)
@asgi_option
@click.option("--browse", is_flag=True, help="open web application in browser")
@click.option("--venv", help="virtual environment location")
@click.argument("nginxfile", type=click.File("rt", encoding="utf-8"))
@click.option(
    "-d",
    "--app-dir",
    "application_dir",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    help="""location of repo or current directory""",
)
def nginx_run_cmd(
    nginxfile: IO[str],
    application_dir: str | None,
    entrypoint: str | None,
    port: int,
    browse: bool,
    venv: str | None,
    asgi: bool,
) -> None:
    """Run nginx as a non daemon process using generated app config file.

    This will test the generated nginx configuration file
    """
    import signal
    import threading
    from tempfile import NamedTemporaryFile

    from ..utils import browser

    nginx_exe = which("nginx")

    def once(m: str) -> Callable[[re.Match[str]], str]:
        done = False

        def f(r: re.Match[str]) -> str:
            nonlocal done
            if done:
                return ""
            done = True
            return m

        return f

    def get_server() -> tuple[str, str | None]:
        """parse nginx.conf file for server and host"""

        def tohost(h: str) -> str | None:
            if h.startswith("unix:"):
                return None
            return h

        A = re.compile("access_log [^;]+;")
        L = re.compile("listen [^;]+;")
        H = re.compile(r"proxy_pass\s+http://([^/\s]+)/?\s*;")
        S = re.compile(r"server\s+([^{\s]+)/?.*;")

        server = nginxfile.read()
        # remove old access_log and replace listen commands
        server = A.sub("", server)
        server = L.sub(once(f"listen {port};"), server)

        m = S.search(server) or H.search(server)

        return server, None if not m else tohost(m.group(1))

    template: Template = get_template("nginx-app.conf", application_dir)
    server, host = get_server()
    application_dir = application_dir or "."

    res = template.render(server=server)  # pylint: disable=no-member
    threads: list[threading.Thread] = []

    with NamedTemporaryFile("w") as fp:
        fp.write(res)
        fp.flush()
        url = f"http://127.0.0.1:{port}"
        click.secho(f"listening on {url}", fg="green", bold=True)
        thrd = None
        bind = "unix:app.sock" if host is None else host
        pidfile = fp.name + ".pid"
        if application_dir:
            if venv is None:
                venv = get_default_venv()

            gunicorn = Path(venv) / "bin" / "gnuicorn"
            if not gunicorn.exists():
                gunicorn = Path(which("gunicorn"))
            entry = entrypoint or get_app_entrypoint(application_dir, asgi=asgi)
            thrd = threading.Thread(
                target=run_app,
                args=[application_dir, str(gunicorn), bind, pidfile, entry],
            )
            # t.setDaemon(True)
            thrd.start()
            threads.append(thrd)
        else:
            entry = entrypoint or get_app_entrypoint(application_dir or ".", asgi=asgi)
            click.secho(
                f"expecting app: gunicorn --bind {bind} {entry}",
                fg="magenta",
            )
        if browse:
            threads.append(browser(url))
        try:
            subprocess.run([nginx_exe, "-c", fp.name], check=False)
        finally:
            if thrd:
                if os.path.isfile(pidfile):
                    with open(pidfile, encoding="utf-8") as fp2:
                        pid = int(fp2.read().strip())
                        os.kill(pid, signal.SIGINT)
            for thrd in threads:
                thrd.join(timeout=2.0)
            rmfiles([pidfile])
            os.system("stty sane")


@config.command(name="nginx-install")
@click.argument(
    "nginxfile",
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
)
def nginx_install_cmd(nginxfile: str) -> None:
    """Install nginx config file."""

    # install frontend
    conf = nginx_install(nginxfile)
    if conf is None:
        raise click.Abort()

    click.secho(f"{conf} installed!", fg="green", bold=True)


@config.command(name="nginx-uninstall")
@click.argument("nginxfile")
def nginx_uninstall_cmd(nginxfile: str) -> None:
    """Uninstall nginx config file."""

    nginx_uninstall(nginxfile)

    click.secho(f"{nginxfile} uninstalled!", fg="green", bold=True)


@config.command(name="nginx-ssl")
@click.option("--days", default=365, help="days of validity")
@click.argument(
    "server_name",
    required=True,
)
def nginx_ssl_cmd(server_name: str, days: int = 365) -> None:
    """Generate openssl TLS self-signed key for a website"""

    ssl_dir = "/etc/ssl"
    openssl = which("openssl")
    sudo = which("sudo")

    country = server_name.split(".")[-1].upper()

    cmd = [
        sudo,
        openssl,
        "req",
        "-x509",
        "-nodes",
        "-days",
        str(days),
        "-newkey",
        "rsa:2048",
        "-keyout" f"{ssl_dir}/private/{server_name}.key" "-out",
        f"{ssl_dir}/certs/{server_name}.crt" "-subj",
        f"/C={country}/CN={server_name}",
    ]

    subprocess.run(cmd, check=True)
    click.secho(f"written keys for {server_name} to {ssl_dir}", fg="green")



from __future__ import annotations

import os
import subprocess
from dataclasses import replace
from typing import NamedTuple

import click

from .cli import cli
from .url import make_url
from .url import URL
from .utils import human
from .utils import which

DB_SIZE = """
SELECT table_name,
    table_rows,
    data_length,
    index_length,
    data_free
FROM information_schema.TABLES
WHERE table_schema = '{db}'
"""


DB_SIZE2 = """
SELECT table_name as "table",
    data_length + index_length as "total_bytes"
FROM information_schema.TABLES
WHERE table_schema = '{db}'
"""
DB_SIZE3 = """
SELECT
    sum(data_length + index_length) as "total_bytes"
FROM information_schema.TABLES
WHERE table_schema = '{db}'
"""


class MySQL:
    def __init__(self, host: str | None = None):
        self.host = host
        self._url: URL | None = None

    @property
    def url(self) -> URL:
        if self._url is not None:
            return self._url
        if self.host is None:
            raise click.BadOptionUsage("host", "please specify HOST")
        self._url = make_url(self.host)
        if self._url is None:
            raise click.BadOptionUsage("host", f"can't parse {self.host}")
        return self._url


def ensure_url(url: str | URL) -> URL:
    ret = make_url(url)
    if ret is None:
        raise click.BadOptionUsage("host", f"can't parse {url}")
    return ret


class Dbsize(NamedTuple):
    table_name: str
    table_rows: int
    data_length: int
    index_length: int
    data_free: int

    @property
    def total(self) -> int:
        return self.data_length + self.index_length


class MySQLError(RuntimeError):
    pass


def mysql_cmd(mysqlexe: str, db: URL, nodb: bool = False) -> list[str]:
    cmd = [mysqlexe]
    if db.username is not None:
        cmd.append(f"--user={db.username}")
    if db.password is not None:
        cmd.append(f"--password={db.password}")
    else:
        cmd.append("-p")
    if db.port is not None:
        cmd.append(f"--port={db.port}")
    if db.host:
        cmd.append(f"--host={db.host}")
    if db.database and not nodb:
        cmd.append(db.database)
    return cmd


def waitfor(procs: list[subprocess.Popen[bytes]]) -> bool:
    ok = True
    for p in procs:
        returncode = p.wait()
        if returncode != 0:
            ok = False
    return ok


class MySQLRunner:
    def __init__(
        self,
        url: str | URL,
        cmds: list[str] | None = None,
        mysqlcmd: str = "mysql",
    ):
        self.url = ensure_url(url)
        mysqlcmd = which(mysqlcmd)
        self.mysql = mysqlcmd
        self.cmds = cmds

    def run(self, query: str | None, nodb: bool = False) -> list[list[str]]:
        db = self.url

        cmd = mysql_cmd(self.mysql, db, nodb=nodb)
        if self.cmds is not None:
            cmd = cmd + self.cmds
        p = subprocess.Popen(
            cmd,
            # stderr=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            text=True,
        )
        stdout, stderr = p.communicate(query)
        if p.returncode != 0:
            stderr = stderr.replace(
                "mysql: [Warning] Using a password on the command line interface can be insecure.",
                "",
            ).strip()
            raise MySQLError(stderr)
        ret: list[list[str]] = []
        for line in stdout.splitlines():
            lines = line.split("\t")
            ret.append(lines)
        return ret


def db_size(url: str | URL, tables: list[str] | None = None) -> int:
    runner = MySQLRunner(url)

    query = DB_SIZE2.format(db=runner.url.database)
    ret = runner.run(query)

    total = 0
    for name, num_bytes in ret[1:]:
        if tables is not None and name not in tables:
            continue
        total += int(num_bytes)
    return total


def db_size_full(
    url: str | URL,
    tables: list[str] | None = None,
) -> list[Dbsize]:
    runner = MySQLRunner(url)

    query = DB_SIZE.format(db=runner.url.database)
    ret = runner.run(query)
    # rows,bytes,index,total, free
    r: list[Dbsize] = []
    for row in ret[1:]:
        name = row[0]
        if tables is not None and name not in tables:
            continue
        vals = [int(r) for r in row[1:]]

        r.append(Dbsize(name, *vals))
    return r


def get_db(url: str | URL) -> list[str]:
    runner = MySQLRunner(url)
    ret = runner.run("show databases", nodb=True)
    return [r[0] for r in ret[1:]]


def get_tables(url: str | URL) -> list[str]:
    runner = MySQLRunner(url)
    ret = runner.run("show tables")
    return [r[0] for r in ret[1:]]


def mysqlload(
    url_str: str | URL,
    filename: str,
    drop: bool = False,
    database: str | None = None,
) -> tuple[int, int]:
    url = ensure_url(url_str)

    if database is not None:
        url = replace(url, database=database)
    if url.database is None:
        raise ValueError(f"no database {url_str}")
    zcat = which("zcat")
    mysqlcmd = which("mysql")

    filesize = os.stat(filename).st_size

    r = MySQLRunner(url)
    if drop:
        r.run(f"drop database if exists {url.database}", nodb=True)
    r.run(
        f"create database if not exists {url.database} character set=latin1",
        nodb=True,
    )

    pzcat = subprocess.Popen(
        [zcat, filename],
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
    )

    cmd = mysql_cmd(mysqlcmd, url)

    pmysql = subprocess.Popen(cmd, stdin=pzcat.stdout, stderr=subprocess.DEVNULL)
    if pzcat.stdout is not None:
        pzcat.stdout.close()

    # pmysql.communicate()
    if not waitfor([pmysql, pzcat]):
        raise MySQLError(f"failed to load {filename}")

    size = db_size(url)

    return size, filesize


def mysqldump(
    url_str: str | URL,
    directory: str | None = None,
    with_date: bool = False,
    tables: list[str] | None = None,
    postfix: str = "",
    database: str | None = None,
) -> tuple[int, int, str]:
    from datetime import datetime
    from .utils import rmfiles
    from pathlib import Path

    url = ensure_url(url_str)
    if database is not None:
        url = replace(url, database=database)
    mysqldumpcmd = which("mysqldump")
    gzip = which("gzip")

    if postfix and not postfix.startswith("-"):
        postfix = "-" + postfix

    if with_date:
        now = datetime.now()
        outname = (
            f"{url.database}{postfix}-{now.year}-{now.month:02}-{now.day:02}.sql.gz"
        )
    else:
        outname = f"{url.database}{postfix}.sql.gz"

    directory = directory or "."

    pth = Path(directory)

    if not pth.exists():
        pth.mkdir(parents=True, exist_ok=True)

    outpath = pth / outname

    cmds = mysql_cmd(mysqldumpcmd, url)
    cmds.extend(["--max_allowed_packet=32M", "--single-transaction"])
    if tables:
        cmds.extend(tables)

    with outpath.open("wb") as fp:
        pmysql = subprocess.Popen(
            cmds,
            stderr=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
        )
        pgzip = subprocess.Popen(
            [gzip],
            stdin=pmysql.stdout,
            stderr=subprocess.DEVNULL,
            stdout=fp,
        )
        if pmysql.stdout is not None:
            pmysql.stdout.close()

    if not waitfor([pmysql, pgzip]):
        rmfiles([str(outpath)])
        raise MySQLError(f"failed to dump database {url.database}")

    filesize = outpath.stat().st_size

    total_bytes = db_size(url, tables)

    return total_bytes, filesize, outname


def analyze(url: URL) -> list[list[str]]:
    tables = ",".join(get_tables(url))
    runner = MySQLRunner(url)

    return runner.run(f"analyze table {tables}")


def tabulate(result: list[list[str]]) -> None:
    def pad(val: str, length: int) -> str:
        p = " " * (length + 1 - len(val))
        return f"{val}{p}"

    if not result:
        return
    max_lengths = [0] * len(result[0])

    for row in result:
        lengths = [len(r) for r in row]
        max_lengths = [max(l1, l2) for l1, l2 in zip(lengths, max_lengths)]

    for idx, row in enumerate(result):
        row = [pad(v, ml) for v, ml in zip(row, max_lengths)]
        print(" ".join(row))
        if idx == 0:
            row = ["=" * (n + 1) for n in max_lengths]
            print(" ".join(row))


def totables(url: URL, tables: tuple[str, ...]) -> list[str] | None:
    only = [t.strip() for tt in tables for t in tt.split(",") if t.strip()]
    if not only:
        return None

    unknown = set(only) - set(get_tables(url))

    if unknown:
        raise click.BadParameter(
            f"unknown table name(s): {' '.join(unknown)}",
            param_hint="tables",
        )
    return only


pass_mysql = click.make_pass_decorator(MySQL)


@cli.group(help=click.style("mysql dump/load commands", fg="magenta"))
@click.option(
    "-h",
    "--host",
    metavar="HOST",
    help="database URL [envvar=HOST]",
    envvar="HOST",
)
@click.pass_context
def mysql(ctx: click.Context, host: str | None) -> None:
    ctx.obj = MySQL(host)


@mysql.command(name="db-size")
@click.option("-s", "--summary", is_flag=True, help="show database total only")
@click.option("-t", "--tables", help="comma separated list of tables", multiple=True)
@click.option("-b", "--bytes", "asbytes", is_flag=True, help="output bytes")
@click.option("-d", "--database", help="database to use (instead of url)")
@pass_mysql
def db_size_cmd(
    db: MySQL,
    tables: tuple[str, ...],
    asbytes: bool,
    summary: bool,
    database: str | None,
) -> None:
    """Print the database size."""
    url = db.url
    if database is not None:
        url = replace(url, database=database)

    only = totables(url, tables)

    if summary:
        total = db_size(url, only)
        click.echo(str(total) if asbytes else human(total))
    else:
        ret = db_size_full(url, only)
        ret = sorted(ret, key=lambda t: -t.total)
        tot = Dbsize(
            table_name="Total",
            table_rows=sum(r.table_rows for r in ret),
            data_length=sum(r.data_length for r in ret),
            index_length=sum(r.index_length for r in ret),
            data_free=sum(r.data_free for r in ret),
        )
        ret.append(tot)
        mx = max(len(r.table_name) for r in ret)

        for r in ret:
            pad = " " * (mx - len(r.table_name))
            v = str(r.total) if asbytes else human(r.total)
            click.echo(f"{r.table_name} {pad}: {v}")


@mysql.command()
@pass_mysql
def databases(db: MySQL) -> None:
    """List databases from URL."""
    for database in sorted(get_db(db.url)):
        print(database)


@mysql.command(name="analyze")
@click.option("-d", "--database", help="database to use (instead of url)")
@pass_mysql
def analyze_cmd(db: MySQL, database: str | None) -> None:
    """Run `analyze table` over database"""
    rurl = db.url
    if database is not None:
        rurl = replace(rurl, database=database)
    tabulate(analyze(rurl))


@mysql.command(name="load")
@click.option("--drop", is_flag=True, help="drop existing database first")
@click.option("-d", "--database", help="database to use (instead of url)")
@click.argument(
    "filename",
    type=click.Path(dir_okay=False, file_okay=True, exists=True),
)
@pass_mysql
def mysqload_cmd(db: MySQL, filename: str, drop: bool, database: str | None) -> None:
    """Load a mysqldump."""

    total_bytes, filesize = mysqlload(db.url, filename, drop=drop, database=database)
    click.secho(
        f"loaded {human(filesize)} > {human(total_bytes)} from {filename}",
        fg="green",
        bold=True,
    )


@mysql.command(name="dump")
@click.option("-p", "--postfix", help="postfix this to database name", default="")
@click.option("--with-date", is_flag=True, help="add a date stamp to filename")
@click.option("-t", "--tables", help="comma separated list of tables", multiple=True)
@click.option("-d", "--database", help="database to use (instead of url)")
@click.argument("directory", required=False)
@pass_mysql
def mysqldump_cmd(
    db: MySQL,
    directory: str | None,
    with_date: bool,
    postfix: str,
    tables: tuple[str, ...],
    database: str | None,
) -> None:
    """Generate a mysqldump to a directory."""
    url = db.url
    if database is not None:
        url = replace(url, database=database)

    tbls = totables(url, tables)

    total_bytes, filesize, outname = mysqldump(
        url,
        directory,
        with_date=with_date,
        tables=tbls,
        postfix=postfix,
        database=database,
    )
    click.secho(
        f"dumped {human(total_bytes)} > {human(filesize)} as {outname}",
        fg="green",
        bold=True,
    )


@mysql.command("query")
@click.argument("query", required=False)
@pass_mysql
def query_cmd(
    db: MySQL,
    query: str | None,
) -> None:
    """Run a query on a mysql database"""
    import sys

    if query is None:
        query = sys.stdin.read()
    runner = MySQLRunner(db.url)
    result = runner.run(query)
    tabulate(result)



from __future__ import annotations

import os
import re
from contextlib import redirect_stderr
from io import StringIO
from os.path import isdir
from os.path import isfile
from typing import Any
from typing import Iterator
from typing import TYPE_CHECKING

import click

from .utils import fixstatic
from .utils import get_dot_env
from .utils import StaticFolder
from .utils import topath


if TYPE_CHECKING:
    from flask import Flask
    from werkzeug.routing import Rule


# core ability


STATIC_RULE = re.compile("^(.*)/<path:filename>$")


def get_flask_static_folders(app: Flask) -> list[StaticFolder]:  # noqa: C901

    def get_static_folder(rule: Rule) -> str | None:
        bound_method = app.view_functions[rule.endpoint]
        if hasattr(bound_method, "static_folder"):
            return getattr(bound_method, "static_folder")
        # __self__ is the blueprint of send_static_file method
        if hasattr(bound_method, "__self__"):
            bp = getattr(bound_method, "__self__")
            if bp.has_static_folder:
                return bp.static_folder
        # now just a lambda :(
        return None

    def find_static(app: Flask) -> Iterator[StaticFolder]:
        if app.has_static_folder:
            prefix, folder = app.static_url_path, app.static_folder
            if folder is not None and isdir(folder):
                yield StaticFolder(
                    prefix,
                    topath(folder),
                    (not folder.endswith(prefix) if prefix else False),
                )
        for r in app.url_map.iter_rules():
            if not r.endpoint.endswith("static"):
                continue
            m = STATIC_RULE.match(r.rule)
            if not m:
                continue
            rewrite = False
            prefix = m.group(1)
            folder = get_static_folder(r)
            if folder is None:
                if r.endpoint != "static":
                    # static view_func for app is now
                    # just a lambda.
                    click.secho(
                        f"location: can't find static folder for endpoint: {r.endpoint}",
                        fg="red",
                        err=True,
                    )
                continue
            if not folder.endswith(prefix):
                rewrite = True

            if not isdir(folder):
                continue
            yield StaticFolder(prefix, topath(folder), rewrite)

    return list(find_static(app))


def is_flask_app(app: Any) -> bool:
    try:
        try:
            # flask and quart obey these
            from flask.sansio.app import App  # type: ignore

            return isinstance(app, App)
        except ModuleNotFoundError:
            from flask import Flask

            return isinstance(app, Flask)
    except ImportError:
        return False


def get_static_folders_for_app(
    application_dir: str,
    entrypoint: str,
    *,
    prefix: str = "",
) -> list[StaticFolder]:
    from .asgi import get_starlette_static_folders
    from .asgi import is_starlette_app

    app = find_application(
        entrypoint,
        application_dir,
    )

    if is_flask_app(app):  # only place we need flask
        return [fixstatic(s, prefix) for s in get_flask_static_folders(app)]
    elif is_starlette_app(app):
        return [fixstatic(s, prefix) for s in get_starlette_static_folders(app)]
    raise click.BadParameter(
        f"{app} is not a flask, quart, starlette or fastapi application!",
    )


def find_application(module: str, application_dir: str | None = None) -> Any:
    import sys
    from importlib import import_module
    from click import style

    remove = False

    if ":" in module:
        module, attr = module.split(":", maxsplit=1)
    else:
        attr = "application"
    if application_dir and application_dir not in sys.path:
        sys.path.append(application_dir)
        remove = True
    try:
        # FIXME: we really want to run this
        # under the virtual environment that this pertains too
        venv = sys.prefix
        click.secho(
            f"trying to load application ({module}) using {venv}: ",
            fg="yellow",
            nl=False,
            err=True,
        )
        with redirect_stderr(StringIO()) as stderr:
            m = import_module(module)
            app: Any = m
            for attr_str in attr.split("."):
                app = getattr(app, attr_str, None)
                if app is None:
                    raise click.BadParameter(
                        f"{attr} doesn't exist for module {module}",
                    )
        v = stderr.getvalue()
        if v:
            click.secho(f"got possible errors ...{style(v[-100:], fg='red')}", err=True)
        else:
            click.secho("ok", fg="green", err=True)

        return app
    except (ImportError, AttributeError) as e:
        raise click.BadParameter(
            f"can't load application from {application_dir}: {e}",
        ) from e
    finally:
        if remove:
            assert application_dir is not None
            sys.path.remove(application_dir)


def get_app_entrypoint(
    application_dir: str,
    *,
    asgi: bool,
    default: str = "app.app:application",
) -> str:
    if asgi:
        ENVS = ["QUART_APP", "FASTAPI_APP", "UVICORN_APP"]
        dotenvs = [".quartenv", ".fastapienv", ".env"]
    else:
        ENVS = ["FLASK_APP"]
        dotenvs = [".flaskenv", ".env"]
    for e in ENVS:
        app = os.environ.get(e)
        if app is not None:
            return app
    for dotenv in dotenvs:
        dot = os.path.join(application_dir, dotenv)
        if isfile(dot):
            cfg = get_dot_env(dot)
            if cfg is None:
                continue
            for e in ENVS:
                app = cfg.get(e)
                if app is not None:
                    return app
    return default



from __future__ import annotations

from . import irds as irds
from . import mailer as mailer
from . import mysql as mysql
from . import remote as remote
from . import restartd as restartd
from . import rsync as rsync
from . import watch as watch
from .cli import cli
from .systemd import nginx as nginx
from .systemd import supervisor as supervisor
from .systemd import systemd as systemd


if __name__ == "__main__":
    cli.main(prog_name="footprint")



from __future__ import annotations

import smtplib
from email.mime.text import MIMEText

import click

from .cli import cli

# from email.mime.image import MIMEImage
# from email.mime.multipart import MIMEMultipart


def sendmail(
    html: str,
    you: str,
    me: str = "footprint@uwa.edu.au",
    mailhost: str | None = None,
    subject: str = "footprint monitor",
    timeout: float = 20.0,
) -> None:
    from .config import get_config

    if mailhost is None:
        mailhost = get_config().mailhost
    msg = MIMEText(html, "html")

    msg["Subject"] = subject
    msg["From"] = me
    msg["To"] = you

    with smtplib.SMTP(timeout=timeout) as s:
        s.connect(mailhost)
        s.sendmail(me, [you], msg.as_string())


@cli.command()
@click.option("-m", "--mailhost", help="mail host to use [default from config]")
@click.option("-t", "--timeout", default=20.0, help="timeout to wait for connection")
@click.argument("email")
@click.argument("message", nargs=-1)
def email_test(
    email: str,
    message: list[str],
    mailhost: str | None,
    timeout: float,
) -> None:
    """Test email setup from this host"""
    import platform

    if not message:
        raise click.BadArgumentUsage("no message")

    message = [*message, f" (Sent via {mailhost})"]

    sendmail(
        " ".join(message),
        you=email,
        mailhost=mailhost,
        subject=f"Message from footprint on {platform.node()}",
        timeout=timeout,
    )
    click.secho("message sent!", fg="green", bold=True)



from __future__ import annotations

import getpass
import math
import os
import subprocess
from contextlib import contextmanager
from contextlib import suppress
from dataclasses import dataclass
from os.path import abspath
from os.path import expanduser
from os.path import normpath
from shutil import which as shwitch
from threading import Thread
from typing import Any
from typing import Iterator
from typing import TYPE_CHECKING
from typing import TypeVar

import click

if TYPE_CHECKING:
    from jinja2 import Template


@dataclass
class StaticFolder:
    url: str | None
    folder: str
    rewrite: bool  # use nginx `rewrite {{url}}/(.*) /$1 break;``


def fixstatic(s: StaticFolder, prefix: str) -> StaticFolder:
    url = prefix + (s.url or "")
    if url and s.folder.endswith(url):
        path = s.folder[: -len(url)]
        return StaticFolder(url, path, False)
    return StaticFolder(url, s.folder, s.rewrite if not prefix else True)


def topath(path: str) -> str:
    return normpath(abspath(expanduser(path)))


def get_dot_env(fname: str) -> dict[str, str | None] | None:
    try:
        from dotenv import dotenv_values  # type: ignore

        return dotenv_values(fname)
    except ImportError:
        import click

        click.secho(
            '".flaskenv" file detected but no python-dotenv module found',
            fg="yellow",
            bold=True,
            err=True,
        )
        return None


def human(num: int, suffix: str = "B", scale: int = 1) -> str:
    if not num:
        return f"0{suffix}"
    num *= scale
    magnitude = int(math.floor(math.log(abs(num), 1000)))
    val = num / math.pow(1000, magnitude)
    if magnitude > 7:
        return f"{val:.1f}Y{suffix}"
    mag = ("", "k", "M", "G", "T", "P", "E", "Z")[magnitude]
    return f"{val:3.1f}{mag}{suffix}"


def rmfiles(files: list[str]) -> None:
    for f in files:
        with suppress(OSError):
            os.remove(f)


def get_pass(VAR: str, msg: str) -> str:
    if VAR not in os.environ:
        return getpass.getpass(f"{msg} password: ")
    return os.environ[VAR]


def multiline_comment(comment: str) -> list[str]:
    return [f"// {line}" for line in comment.splitlines()]


def flatten_toml(d: dict[str, Any]) -> dict[str, Any]:
    def inner(
        d: dict[str, Any],
        view: str = "",
        level: int = 0,
    ) -> Iterator[tuple[str, Any]]:
        for k, v in d.items():
            if "." in k:
                continue
            if isinstance(v, dict) and level == 0:
                yield from inner(v, f"{view}{k}.", level=level + 1)  # type: ignore
            else:
                yield f"{view}{k}", v

    return dict(inner(d))


def gethomedir(user: str = "") -> str:
    user = user.replace("\\\\", "\\")
    return os.path.expanduser(f"~{user}")


def toml_load(path: str) -> dict[str, Any]:
    try:
        import tomllib  # type: ignore

        with open(path, "rb") as fp:
            return tomllib.load(fp)
    except ImportError:
        import toml  # type: ignore

        return toml.load(path)


def browser(url: str = "http://127.0.0.1:2048", sleep: float = 2.0) -> Thread:
    import time
    import webbrowser

    def run() -> None:
        time.sleep(sleep)
        webbrowser.open_new_tab(url)

    tr = Thread(target=run)
    tr.start()
    return tr


@dataclass
class Runner:
    name: str
    cmd: list[str]
    directory: str
    warn: bool = True
    showcmd: bool = False
    env: dict[str, str] | None = None
    shell: bool = False

    def run(self) -> subprocess.Popen[bytes]:
        click.secho(f"starting {self.name}", fg="yellow")
        if self.showcmd:
            click.echo(" ".join(str(s) for s in self.cmd))
        ret = subprocess.Popen(
            self.cmd,
            cwd=self.directory,
            env=self.getenv(),
            shell=self.shell,
        )
        return ret

    def getenv(self) -> dict[str, str] | None:
        if not self.env:
            return None
        return {**os.environ, **self.env}

    def start(self) -> subprocess.Popen[bytes]:
        return self.run()


def is_local(machine: str | None) -> bool:
    return machine in {None, "127.0.0.1", "localhost"}


T = TypeVar("T")


@contextmanager
def maybe_closing(thing: T) -> Iterator[T]:
    try:
        yield thing
    finally:
        if hasattr(thing, "close"):
            thing.close()  # type: ignore


def userdir() -> str:
    pth = os.environ.get("XDG_CONFIG_HOME")
    if pth:
        return os.path.join(pth, "systemd", "user")
    return os.path.expanduser("~/.config/systemd/user")


def get_variables(template: Template) -> set[str]:
    from jinja2 import meta

    if template.filename is None or template.filename == "<template>":
        return set()
    env = template.environment
    with open(template.filename, encoding="utf-8") as fp:
        ast = env.parse(fp.read())
    return meta.find_undeclared_variables(ast)


def which(cmd: str) -> str:
    ret = shwitch(cmd)
    if ret is None:
        click.secho(f"no executable {cmd}!", fg="red", err=True)
        raise click.Abort()
    return ret


def require_mod(mod: str, mod_name: str | None = None) -> None:
    from importlib import import_module

    try:
        import_module(mod)
        return
    except ModuleNotFoundError as e:
        import sys

        if mod_name is None:
            mod_name = mod
        click.secho(
            f"Please install {mod} ({sys.executable} -m pip install {mod_name})",
            fg="yellow",
            bold=True,
            err=True,
        )
        raise click.Abort() from e



from __future__ import annotations

import os
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields
from dataclasses import replace
from typing import IO

from .utils import toml_load


REPO = "git+https://github.com/arabidopsis/footprint.git"


@dataclass
class Config:
    mailhost: str = "uwa-edu-au.mail.protection.outlook.com"
    # mailhost: str = "antivirus.uwa.edu.au"
    datastore: str = "//drive.irds.uwa.edu.au/sci-ms-001"
    # directories that *might* be in the static directory
    static_dir: list[str] = field(
        default_factory=lambda: [
            "img",
            "images",
            "js",
            "css",
            "media",
            "docs",
            "tutorials",
            "notebooks",
            "downloads",
            ".well-known",
        ],
    )
    # basic files that have urls such as /robots.txt /favicon.ico etc.
    static_files: list[str] = field(
        default_factory=lambda: [
            "robots.txt",
            "crossdomain.xml",
            "favicon.ico",
            "browserconfig.xml",
            "humans.txt",
        ],
    )
    # exclude these filenames/directories from static consideration
    exclude: list[str] = field(default_factory=lambda: ["__pycache__"])
    # directory to put config files: (Ubuntu, RHEL8)
    nginx_dirs: list[str] = field(
        default_factory=lambda: ["/etc/nginx/sites-enabled", "/etc/nginx/conf.d"],
    )
    arg_color: str = "yellow"  # use "none" for no color


XConfig: Config | None = None


def get_config() -> Config:
    global XConfig
    if XConfig is None:
        XConfig = _init_config(Config())
    return XConfig


def _init_config(config: Config, application_dir: str = ".") -> Config:
    project = os.path.join(application_dir, "pyproject.toml")
    if os.path.isfile(project):
        try:
            d = toml_load(project)
            if "tool" not in d:
                return config
            cfg = d["tool"].get("footprint")
            if cfg is None:
                return config
            data = {}
            for f in fields(config):
                if f.name in cfg:
                    data[f.name] = cfg[f.name]

            if data:
                config = replace(config, **data)

        except ImportError:
            pass
        except Exception:
            import click

            click.secho(f'can\'t load "{project}"', fg="red", bold=True, err=True)
    return config


def dump_toml(config: Config, out: IO[str]) -> bool:
    try:
        import toml  # type: ignore

        d = dict(tool=dict(footprint=asdict(config)))
        toml.dump(d, out)
        return True
    except Exception:
        return False


def dump_to_file(filename: str, append: bool) -> bool:
    config = get_config()
    if filename == "-":
        import sys

        return dump_toml(config, sys.stdout)
    with open(filename, "a" if append else "w", encoding="utf-8") as fp:
        return dump_toml(config, fp)



from __future__ import annotations

import subprocess

import click

from .cli import cli
from .utils import which


def vmemory_ok(threshold: int = 100) -> list[str]:
    import psutil

    from .utils import human

    m = psutil.virtual_memory()
    mn = threshold * 1024 * 1024  # megabyte
    if mn <= 0:
        return [f"memory available: {human(m.available)} ({m.percent}% used)"]

    if m.available < mn:
        return [f"low memory: {human(m.available)} < {human(mn)} ({m.percent}% used)"]
    return []


def disks_ok(threshold: int = 100) -> list[str]:
    import psutil

    from .utils import human

    mounts = [
        p.mountpoint
        for p in psutil.disk_partitions()
        if not p.device.startswith("/dev/loop") and not p.mountpoint.startswith("/boot")
    ]
    mn = threshold * 1024 * 1024  # megabytes

    ret: list[str] = []
    app = ret.append
    for m in mounts:
        du = psutil.disk_usage(m)
        if mn <= 0:
            app(f"partition {m}: {human(du.free)} Avail ({du.percent}% used)")
        elif du.free < mn:
            app(f"partition {m}: {human(du.free)} < {human(mn)} ({du.percent}% used)")
    return ret


# @cli.command()
# @watch_options
# @click.argument("email", required=False)
def run_watch(
    email: str | None,
    mem_threshold: int,
    disk_threshold: int,
    mailhost: str,
) -> None:
    import platform

    if disk_threshold > 0 and mem_threshold > 0:
        status = "Low memory"
    else:
        status = "Status"
    machine = platform.node()
    W = """<strong>{status} on {machine}</strong>:<br/>
{disk}"""
    memory = vmemory_ok(mem_threshold)
    disk = disks_ok(disk_threshold)

    if disk or memory:
        from .mailer import sendmail

        m = "<br/>\n".join(disk + memory)
        msg = W.format(disk=m, machine=machine, status=status)
        if email:
            sendmail(msg, email, mailhost=mailhost, subject=f"{status} on {machine}")
        else:
            click.echo(msg)


def add_cron_command(cmd: str, test_line: str | None = None) -> None:
    from tempfile import NamedTemporaryFile

    crontab = which("crontab")

    p = subprocess.run(
        [crontab, "-l"],
        capture_output=True,
        check=False,
        text=True,
    ).stdout

    ct = []
    added = False
    for line in p.splitlines():
        if test_line is None or test_line in line:
            ct.append(cmd)
            added = True
        else:
            ct.append(line)
    if not added:
        ct.append(cmd)

    with NamedTemporaryFile("wt") as fp:
        fp.write("\n".join(ct))
        fp.write("\n")
        fp.flush()
        # load new crontab
        subprocess.run([crontab, fp.name], check=True)


def make_cron_interval(interval_mins: int) -> str:
    if interval_mins >= 60:
        h = int(interval_mins // 60)
        tme = f"0 */{h} * * *"
    else:
        tme = f"*/{interval_mins} * * * *"
    return tme


@cli.command(
    epilog=click.style(
        'Use "crontab -l" to see if watch has been installed',
        fg="magenta",
    ),
)
@click.option(
    "-t",
    "--mem-threshold",
    default=100,
    help="memory min free space in megabytes",
    show_default=True,
)
@click.option(
    "-d",
    "--disk-threshold",
    default=100,
    help="disk partition min free space in megabytes",
    show_default=True,
)
@click.option(
    "-m",
    "--mailhost",
    help="SMTP mail host to connect to",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="send email whatever",
)
@click.option(
    "-i",
    "--interval",
    default=10,
    help="check interval in minutes",
    show_default=True,
)
@click.option("-c", "--crontab", is_flag=True, help="install command into crontab")
@click.option("--test", "is_test", is_flag=True, help="show cron command only")
@click.argument("email", required=False)
def watch(
    email: str,
    crontab: bool,
    mem_threshold: int,
    disk_threshold: int,
    mailhost: str,
    interval: int,
    force: bool,
    is_test: bool,
) -> None:
    """Install a crontab watch on low memory and diskspace [**requires psutil**]"""
    import sys
    from .utils import require_mod
    from .config import get_config

    require_mod("psutil")

    if mailhost is None:
        mailhost = get_config().mailhost

    if force and crontab:
        raise click.BadParameter("can't specifiy --force *and* --crontab")

    if not crontab:
        if force:
            mem_threshold = -1
            disk_threshold = -1
        run_watch(email, mem_threshold, disk_threshold, mailhost)
        return

    if not email:
        raise click.BadArgumentUsage("email must be present if --crontab specified")
    tme = make_cron_interval(interval)

    C = (
        f"{tme} {sys.executable}"
        f" -m footprint watch -m {mailhost} -t {mem_threshold} -d {disk_threshold} {email} 1>/dev/null 2>&1"
    )
    if is_test:
        click.echo(C)
    else:
        add_cron_command(C, "footprint watch")


@cli.command(
    epilog=click.style(
        'Use "crontab -l" to see if watch has been installed',
        fg="magenta",
    ),
)
@click.option(
    "-i",
    "--interval",
    default=10,
    help="check interval in minutes",
    show_default=True,
)
@click.option("-t", "--test", "is_test", is_flag=True, help="show cron command only")
@click.argument("command", nargs=-1)
def cron(command: list[str], interval: int, is_test: bool) -> None:
    """Install a python crontab command"""
    import os
    import sys

    if not command:
        return

    cmd = " ".join(command)
    old = cmd
    tme = make_cron_interval(interval)
    if os.path.isfile(cmd):
        cmd = os.path.abspath(cmd)

    C = f"{tme} {sys.executable} {cmd} 1>/dev/null 2>&1"
    if is_test:
        click.echo(C)
    else:
        add_cron_command(C, old)



from __future__ import annotations

import subprocess
from getpass import getuser

import click

from .cli import cli
from .systemd.systemd import systemd
from .systemd.utils import make_args
from .utils import get_pass
from .utils import which


def mount_irds(path_str: str, user: str | None = None) -> int:
    from .config import get_config
    from pathlib import Path
    import os

    sudo = which("sudo")
    mount = which("mount")

    path = Path(path_str).expanduser().absolute()
    if not path.exists():
        path.mkdir(exist_ok=True, parents=True)

    datastore = path / "datastore"
    if datastore.exists():
        return 0

    if user is None:
        user = getuser()
    uid = os.getuid()
    gid = os.getgid()
    pheme = get_pass("PHEME", f"user {user} pheme")
    cmd = [
        sudo,
        mount,
        "-t",
        "cifs",
        "-o",
        f"user={user}",
        "-o",
        f"pass={pheme}",
        "-o",
        f"uid={uid},gid={gid},forceuid,forcegid",
        get_config().datastore,
        str(path),
    ]
    pmount = subprocess.Popen(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    returncode = pmount.wait()
    return returncode


@cli.group(help=click.style("IRDS commands", fg="magenta"))
def irds() -> None:
    pass


@irds.command(name="mount")
@click.argument("directory")
@click.argument("user", required=False)
def mount_irds_cmd(directory: str, user: str | None) -> None:
    """Mount IRDS datastore."""

    returncode = mount_irds(directory, user=user)
    if returncode != 0:
        click.secho("can't mound irds", fg="red")
        raise click.Abort()


MOUNT_ARGS = {
    "mount_dir": "locations of repo",
    "user": "user to run as [default: current user]",
    "version": "SMB version [default: 3.0]",
    "credentials": "file containg PHEME password as a line: password={pw}"
    " (no spaces)\nroot owned with permission 600",
    "password": "PHEME password",
    "drive": "IRDS drive to mount",
    "uid": "user id",
    "gid": "group id",
}

MOUNT_HELP = f"""
Generate a systemd mount file for a IRDS.

Use footprint irds systemd path/to/mount_dir ... etc.
with the following arguments:

\b
{make_args(MOUNT_ARGS)}
\b
example:
\b
footprint irds systemd ~/irds user=00033472
"""


@irds.command(name="systemd", help=MOUNT_HELP)
@click.option("-i", "--ignore-unknowns", is_flag=True, help="ignore unknown variables")
@click.option(
    "-c",
    "--credentials",
    type=click.Path(file_okay=True, dir_okay=False, exists=True),
    help="credentials file for CIFS",
)
@click.option("-t", "--template", metavar="TEMPLATE_FILE", help="template file")
@click.option("-n", "--no-check", is_flag=True, help="don't check parameters")
@click.argument(
    "mount_dir",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    required=True,
)
@click.argument("params", nargs=-1)
def systemd_mount_cmd(
    mount_dir: str,
    params: list[str],
    template: str | None,
    no_check: bool,
    ignore_unknowns: bool,
    credentials: str | None,
) -> None:
    """Generate a systemd unit file to mount IRDS.

    PARAMS are key=value arguments for the template.
    """
    import os
    from getpass import getpass

    from .config import get_config

    params = list(params)

    Config = get_config()

    mount_dir = mount_dir or "."
    mount_dir = os.path.abspath(os.path.expanduser(mount_dir))

    def isadir(d: str) -> str | None:
        return None if os.path.isdir(d) else f"{d}: not a directory"

    def isafile(d: str) -> str | None:
        return None if os.path.isfile(d) else f"{d}: not a file"

    se = which("systemd-escape")
    filename = subprocess.check_output(
        [se, "-p", "--suffix=mount", mount_dir],
        text=True,
    ).strip()

    if credentials is not None:
        params.append(f"credentials={os.path.expanduser(credentials)}")

    systemd(
        template or "systemd.mount",
        mount_dir,
        params,
        help_args={**MOUNT_ARGS, "port": "unused"},
        check=not no_check,
        output=filename,
        ignore_unknowns=ignore_unknowns,
        checks=[
            (
                "mount_dir",
                lambda _, v: isadir(v),
            ),
            ("credentials", lambda _, v: isafile(v)),
        ],
        default_values=[
            ("uid", lambda _: str(os.getuid())),
            ("gid", lambda _: str(os.getgid())),
            ("drive", lambda _: Config.datastore),
            (
                "password",
                lambda params: (
                    getpass(f"PHEME password for {params['user']}: ")
                    if "credentials" not in params
                    else None
                ),
            ),
        ],
    )
    msg = click.style(f"footprint config systemd-install {filename}", fg="green")
    click.echo(f'use: "{msg}" to install')



from __future__ import annotations

import subprocess
import sys
from os.path import isdir
from os.path import isfile
from os.path import split
from typing import Any
from typing import Callable
from typing import TextIO
from typing import TYPE_CHECKING

import click

from ..core import get_app_entrypoint
from ..core import topath
from ..templating import get_template
from ..templating import undefined_error
from ..utils import get_variables
from ..utils import gethomedir
from ..utils import rmfiles
from ..utils import userdir
from ..utils import which
from .cli import config
from .utils import asgi_option
from .utils import asuser_option
from .utils import check_app_dir
from .utils import check_user
from .utils import check_venv_dir
from .utils import CHECKTYPE
from .utils import config_options
from .utils import CONVERTER
from .utils import fix_params
from .utils import footprint_config
from .utils import get_default_venv
from .utils import get_known
from .utils import getgroup
from .utils import getuser
from .utils import make_args
from .utils import template_option
from .utils import to_check_func
from .utils import to_output

if TYPE_CHECKING:
    from jinja2 import Template


def systemd_install(
    systemdfiles: list[str],  # list of systemd unit files
    asuser: bool = False,  # install as user
) -> list[str]:  # this of failed installations
    import filecmp

    location = userdir() if asuser else "/etc/systemd/system"

    sudo = which("sudo")
    systemctl = which("systemctl")

    def sudocmd(*args: str, check: bool = True) -> subprocess.CompletedProcess[bytes]:
        if not asuser:
            return subprocess.run([sudo] + list(args), check=check)
        return subprocess.run(list(args), check=check)

    def systemctlcmd(*args: str, check: bool = True) -> int:
        if not asuser:
            return subprocess.run(
                [sudo, systemctl] + list(args),
                check=check,
            ).returncode
        return subprocess.run(
            [systemctl, "--user"] + list(args),
            check=check,
        ).returncode

    failed: list[str] = []
    for systemdfile in systemdfiles:
        service = split(systemdfile)[-1]
        exists = isfile(f"{location}/{service}")
        if not exists or not filecmp.cmp(f"{location}/{service}", systemdfile):
            if exists:
                click.secho(f"warning: overwriting old {service}", fg="yellow")

                ret = systemctlcmd("stop", service, check=False)

                if ret != 0:
                    click.secho(
                        "failed to stop old process [already stopped?]",
                        fg="yellow",
                        err=True,
                    )
            # will throw....
            sudocmd("cp", systemdfile, location)
            systemctlcmd("daemon-reload")
            systemctlcmd("enable", service)
            systemctlcmd("start", service)
            if systemctlcmd("status", service):
                systemctlcmd("disable", service, check=False)
                sudocmd("rm", f"{location}/{service}")
                systemctlcmd("daemon-reload")

                click.secho("systemd configuration faulty", fg="red", err=True)
                failed.append(systemdfile)

        else:
            click.secho(f"systemd file {service} unchanged", fg="green")
    return failed


def systemd_uninstall(
    systemdfiles: list[str],
    asuser: bool = False,
) -> list[str]:
    # install systemd file
    location = userdir() if asuser else "/etc/systemd/system"
    sudo = which("sudo")
    systemctl = which("systemctl")

    def sudocmd(*args: str, check: bool = True) -> subprocess.CompletedProcess[bytes]:
        if not asuser:
            return subprocess.run([sudo] + list(args), check=check)
        return subprocess.run(list(args), check=check)

    def systemctlcmd(*args: str, check: bool = True) -> int:
        if not asuser:
            return subprocess.run(
                [sudo, systemctl] + list(args),
                check=check,
            ).returncode
        return subprocess.run(
            [systemctl, "--user"] + list(args),
            check=check,
        ).returncode

    failed: list[str] = []
    changed = False
    for sdfile in systemdfiles:
        systemdfile = split(sdfile)[-1]
        if "." not in systemdfile:
            systemdfile += ".service"
        filename = f"{location}/{systemdfile}"
        if not isfile(filename):
            click.secho(f"no systemd service {systemdfile}", fg="yellow", err=True)
        else:
            ret = systemctlcmd("stop", systemdfile, check=False)
            if ret != 0 and ret != 5:
                failed.append(sdfile)
            if ret == 0:
                systemctlcmd("disable", systemdfile)
                sudocmd("rm", filename)
                changed = True
    if changed:
        systemctlcmd("daemon-reload")
    return failed


SYSTEMD_ARGS = {
    "application_dir": "locations of repo",
    "appname": "application name [default: directory name]",
    "user": "user to run as [default: current user]",
    "group": "group for executable [default: current user's group]",
    "venv": "virtual environment to use [default: {application_dir}/{.venv,../venv}]",
    "workers": "number of gunicorn workers [default: (CPU // 2 + 1) or 2 for ASGI]",
    "stopwait": "seconds to wait for website to stop",
    "after": "start after this service [default: mysql.service]",
    "host": "bind gunicorn to a port [default: use unix socket]",
    "asuser": "systemd destined for --user directory",
    "homedir": "$HOME (default generated from user parameter)",
    "executable": "defaults to sys.executable i.e. the current python",
    "path": "extra bin directories to add to PATH",
    "env-file": "path to a environment file",
}


SYSTEMD_HELP = f"""
Generate a systemd unit file for a website.

Use footprint config systemd ... etc.
with the following arguments:

\b
{make_args(SYSTEMD_ARGS)}
\b
example:
\b
footprint config systemd host=8001
"""


# pylint: disable=too-many-branches too-many-locals
def systemd(  # noqa: C901
    template: str | Template,
    application_dir: str,
    args: list[str] | None = None,
    *,
    help_args: dict[str, str] | None = None,
    check: bool = True,
    output: str | TextIO | None = None,
    extra_params: dict[str, Any] | None = None,
    checks: list[tuple[str, CHECKTYPE]] | None = None,
    asuser: bool = False,
    ignore_unknowns: bool = False,
    default_values: list[tuple[str, CONVERTER]] | None = None,
    convert: dict[str, Callable[[Any], Any]] | None = None,
    asgi: bool = False,
) -> str:
    # pylint: disable=line-too-long
    # see https://www.digitalocean.com/community/tutorials/how-to-serve-flask-applications-with-gunicorn-and-nginx-on-ubuntu-20-04
    # place this in /etc/systemd/system/
    from multiprocessing import cpu_count
    from jinja2 import UndefinedError

    if help_args is None:
        help_args = SYSTEMD_ARGS

    application_dir = topath(application_dir)

    # if not params:
    #     raise click.BadParameter("use --help for params", param_hint="params")
    template = get_template(template, application_dir)
    variables = get_variables(template)
    known: set[str] = (
        get_known(help_args)
        | {"app", "asuser", "asgi"}
        | (set(extra_params.keys()) if extra_params else set())
    )
    known.update(variables)
    defaults: list[tuple[str, CONVERTER]] = [
        ("application_dir", lambda _: application_dir),
        ("asgi", lambda _: asgi),
        ("user", lambda _: getuser()),
        ("group", lambda params: getgroup(params["user"])),
        ("appname", lambda params: split(params["application_dir"])[-1]),
        ("venv", lambda _: get_default_venv()),
        ("homedir", lambda params: gethomedir(params["user"])),
        ("executable", lambda _: sys.executable),
    ]
    if default_values:
        defaults.extend(default_values)
    defaults.extend(
        [
            ("workers", lambda _: 2 if asgi else cpu_count() // 2 + 1),
        ],
    )
    params = {}
    try:
        params = {
            k: v for k, v in footprint_config(application_dir).items() if k in known
        }
        params.update(fix_params(args or [], convert))
        if extra_params:
            params.update(extra_params)

        for key, default_func in defaults:
            if key not in params:
                v = default_func(params)
                if v is not None:
                    params[key] = v
                    known.add(key)

        def isint(s: str | int) -> bool:
            return isinstance(s, int) or s.isdigit()

        if "host" in params:
            h = params["host"]
            if isint(h):
                params["host"] = "127.0.0.1"
                params["port"] = h
            else:
                if ":" in h:
                    s, h = h.rsplit(":", maxsplit=1)
                    params["host"] = s
                    params["port"] = h

        if "port" not in params:
            params["port"] = 8000

        if check:
            if not ignore_unknowns:
                extra = set(params) - known
                if extra:
                    raise click.BadParameter(
                        f"unknown arguments {extra}",
                        param_hint="params",
                    )
            failed: list[str] = []
            checks = list(checks or []) + [
                to_check_func("stopwait", isint, "{stopwait} is not an integer"),
                to_check_func("homedir", isdir, "{homedir} is not a directory"),
            ]
            for key, func in checks:
                if key in params and key:
                    v = params[key]
                    msg = func(key, v)
                    if msg is not None:
                        click.secho(
                            msg,
                            fg="yellow",
                            bold=True,
                            err=True,
                        )
                        failed.append(key)
                if failed:
                    raise click.Abort()

        if "asuser" not in params:
            params["asuser"] = asuser
        if "asgi" not in params:
            params["asgi"] = asgi
        if "app" not in params:
            app = get_app_entrypoint(application_dir, asgi=asgi)
            if asgi and ":" not in app:
                app += ":application"
            params["app"] = app
        res = template.render(**params)  # pylint: disable=no-member
        to_output(res, output)
        return res
    except UndefinedError as e:
        undefined_error(e, template, params)
        raise click.Abort()


def multi_systemd(
    template: str | None,
    application_dir: str | None,
    args: list[str],
    *,
    check: bool = True,
    output: str | None = None,
    asuser: bool = False,
    ignore_unknowns: bool = False,
    asgi: bool = False,
) -> None:
    """Generate a systemd unit file to start gunicorn for this webapp.

    PARAMS are key=value arguments for the template.
    """
    from jinja2 import Template

    from ..templating import get_templates
    from ..utils import maybe_closing

    def get_name(tmpl: str | Template) -> str | None:
        name = tmpl.name if isinstance(tmpl, Template) else output
        name = topath(name) if name else name

        if (
            isinstance(tmpl, Template)
            and name
            and tmpl.filename
            and name == topath(tmpl.filename)
        ):
            raise RuntimeError(f"overwriting template: {name}!")
        return name

    application_dir = application_dir or "."
    templates = get_templates(
        template or ("uvicorn.service" if asgi else "systemd.service"),
    )
    for tmpl in templates:
        name = None
        try:
            name = get_name(tmpl)

            with maybe_closing(
                open(name, "w", encoding="utf-8") if name else None,
            ) as fp:
                systemd(
                    tmpl,
                    application_dir,
                    args,
                    help_args=SYSTEMD_ARGS,
                    check=check,
                    output=fp,
                    asuser=asuser,
                    asgi=asgi,
                    ignore_unknowns=ignore_unknowns,
                    checks=[
                        ("application_dir", lambda _, v: check_app_dir(v)),
                        ("venv", lambda _, v: check_venv_dir(v)),
                    ],
                    convert={"venv": topath, "application_dir": topath},
                )
        except Exception as exc:
            if isinstance(name, str):
                rmfiles([name])
            raise exc


@config.command(name="systemd", help=SYSTEMD_HELP)
@asuser_option
@click.option("-i", "--ignore-unknowns", is_flag=True, help="ignore unknown variables")
@template_option
@config_options
@click.option(
    "-d",
    "--app-dir",
    "application_dir",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    help="""location of repo or current directory""",
)
@asgi_option
@click.argument("params", nargs=-1)
def systemd_cmd(
    application_dir: str | None,
    params: list[str],
    template: str | None,
    no_check: bool,
    output: str | None,
    asuser: bool,
    asgi: bool,
    ignore_unknowns: bool,
) -> None:
    """Generate a systemd unit file to start gunicorn or uvicorn for this webapp.

    PARAMS are key=value arguments for the template.
    """
    from ..utils import require_mod

    if asgi:
        require_mod("uvicorn")
    else:
        require_mod("gunicorn")

    multi_systemd(
        template,
        application_dir,
        params,
        check=not no_check,
        output=output,
        ignore_unknowns=ignore_unknowns,
        asuser=asuser,
        asgi=asgi,
    )


TUNNEL_ARGS = {
    "local-port": "local port to connect to",
    "remote-port": "remote port to connect to",
    "keyfile": "ssh keyfile to use for target machine",
    "remote-user": "remote user to run as [default: current user]",
    "restart": "seconds to wait for before restart [default: 5]",
    "local-addr": "local address to connect [default: 127.0.0.1]",
}
TUNNEL_HELP = f"""
Generate a systemd unit file for a ssh tunnel.

Use footprint config tunnel machine ... etc.
with the following arguments:

\b
{make_args(TUNNEL_ARGS)}
\b
example:
\b
footprint config ssh-tunnel machine1 local-port=8001 remote-port=80
"""


@config.command(name="ssh-tunnel", help=TUNNEL_HELP)
@asuser_option
@click.option("-i", "--ignore-unknowns", is_flag=True, help="ignore unknown variables")
@template_option
@config_options
@click.argument(
    "target",
    required=True,
)
@click.argument("params", nargs=-1)
def tunnel_cmd(
    target: str,
    params: list[str],
    template: str | None,
    no_check: bool,
    output: str | None,
    asuser: bool,
    ignore_unknowns: bool,
) -> None:
    """Generate a systemd unit file to start ssh tunnel to TARGET.

    PARAMS are key=value arguments for the template.
    """
    systemd(
        template or "secure-tunnel.service",
        ".",
        params,
        help_args=TUNNEL_ARGS,
        check=not no_check,
        output=output,
        asuser=asuser,
        extra_params={"target": target},
        ignore_unknowns=ignore_unknowns,
        checks=[
            (
                "keyfile",
                lambda _, f: None if isfile(f) else f'keyfile "{f}" is not a file',
            ),
            (
                "restart",
                lambda _, n: None if n > 2 else "restart {n} is too short an interval",
            ),
        ],
        default_values=[
            ("local_addr", lambda _: "127.0.0.1"),
            ("restart", lambda _: 5),
            ("remote_user", lambda params: params["user"]),
        ],
        convert={"keyfile": topath},
    )


@config.command(name="template")
@asuser_option
@click.option(
    "-o",
    "--output",
    help="write to this file",
    type=click.Path(dir_okay=False),
)
@click.argument(
    "template",
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
)
@click.argument("params", nargs=-1)
def template_cmd(
    params: list[str],
    template: str,
    output: str | None,
    asuser: bool,
) -> None:
    """Generate file from a jinja template.

    PARAMS are key=value arguments for the template.
    """
    systemd(
        template,
        ".",
        params,
        help_args={},
        check=False,
        output=output,
        asuser=asuser,
        ignore_unknowns=True,
    )


@config.command(name="systemd-install")
@asuser_option
@click.argument(
    "systemdfiles",
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
    nargs=-1,
    required=True,
)
def systemd_install_cmd(systemdfiles: list[str], asuser: bool) -> None:
    """Install systemd files."""

    check_user(asuser)

    failed = systemd_install(systemdfiles, asuser=asuser)

    if failed:
        raise click.Abort()


@config.command(name="systemd-uninstall")
@asuser_option
@click.argument(
    "systemdfiles",
    # type=click.Path(exists=True, dir_okay=False, file_okay=True),
    nargs=-1,
    required=True,
)
def systemd_uninstall_cmd(systemdfiles: list[str], asuser: bool) -> None:
    """Uninstall systemd files."""
    check_user(asuser)
    failed = systemd_uninstall(systemdfiles, asuser=asuser)
    if failed:
        click.secho(f'failed to stop: {",".join(failed)}', fg="red", err=True)
        raise click.Abort()



from __future__ import annotations

from os.path import isdir
from os.path import join
from typing import Any
from typing import TextIO
from typing import TYPE_CHECKING

import click

from .cli import config
from .utils import asuser_option
from .utils import CHECKTYPE
from .utils import config_options
from .utils import CONVERTER
from .utils import make_args
from .utils import template_option


if TYPE_CHECKING:
    from jinja2 import Template

SUPERVISORD_ARGS = {
    "application_dir": "locations of all repo",
    "appname": "application name [default: directory name]",
    "annotator": "annotator repo directory",
    "venv": "virtual env directory [default: where python executable exists]",
    "user": "user to run as [default: current user]",
    "group": "group to run as [default: current user group]",
    "workers": "number of julia and celery workers to start [default: 4]",
    "threads": "number of julia threads to use [default: 8]",
    "stopwait": "seconds to wait for julia and celery to stop [default: 30]",
    "heatbeat": "celery worker heatbeat interval in seconds [default: 30]",
    "gevent": "run celery worker with gevent `-P gevent`",
    "max_interval": "interval between beats [default: 3600]",
    "after": "start after this service [default: mysql.service]",
    "celery": "celery --app to start [default: {appname}.celery]",
    "julia": "julia directory",
    "depot_path": "where downloaded julia packages are stored [default: /home/{user}/.julia ]",
}
SUPERVISORD_HELP = f"""
Generate a supervisord conf file for website background.

Use footprint config supervisord ... etc.
with the following params:

\b
{make_args(SUPERVISORD_ARGS)}
\b
example:
\b
footprint config supervisord venv=/home/ianc/miniconda3
"""
CELERY_SYSTEMD_HELP = f"""
Generate a systemd conf file for website background.

Use footprint config systemd-celery ... etc.
with the following params:

\b
{make_args(SUPERVISORD_ARGS)}
\b
example:
\b
footprint config systemd-celery venv=/home/ianc/miniconda3
"""


# pylint: disable=too-many-branches too-many-locals
def supervisor(  # noqa: C901
    template: str | Template,
    application_dir: str | None = None,
    args: list[str] | None = None,
    *,
    help_args: dict[str, str] | None = None,
    check: bool = True,
    output: str | TextIO | None = None,
    extra_params: dict[str, Any] | None = None,
    checks: list[tuple[str, CHECKTYPE]] | None = None,
    ignore_unknowns: bool = False,
    asuser: bool = False,
    default_values: list[tuple[str, CONVERTER]] | None = None,
) -> str:
    import os

    from .systemd import systemd
    from ..core import topath

    def isadir(key: str, s: Any) -> str | None:
        if not isdir(s):
            return f"{key}: {s} is not a directory"
        return None

    def is_julia(key: str, s: Any) -> str | None:
        if not isdir(s):
            return f"{key}: {s} is not a directory"
        if not os.access(join(s, "bin", "julia"), os.X_OK | os.R_OK):
            return f"{key}: {s} is not a *julia* directory"
        return None

    schecks: list[tuple[str, CHECKTYPE]] = [
        ("julia_dir", is_julia),
        ("depot_path", isadir),
    ]
    schecks.extend(checks or [])

    defaults: list[tuple[str, CONVERTER]] = [
        ("depot_path", lambda params: f'{params["homedir"]}/.julia'),
        ("workers", lambda _: 4),
        ("gevent", lambda _: False),
        ("stopwait", lambda _: 10),
    ]
    if default_values:
        defaults = [*default_values, *defaults]

    return systemd(
        template,
        application_dir or ".",
        args,
        help_args=help_args or SUPERVISORD_ARGS,
        check=check,
        output=output,
        asuser=asuser,
        extra_params=extra_params,
        default_values=defaults,
        ignore_unknowns=ignore_unknowns,
        checks=schecks,
        convert={"julia_dir": topath, "depot_path": topath},
    )


def supervisord(
    template: str | None,
    application_dir: str | None,
    args: list[str],
    *,
    help_args: dict[str, str] | None = None,
    check: bool = True,
    output: str | TextIO | None = None,
    extra_params: dict[str, Any] | None = None,
    checks: list[tuple[str, CHECKTYPE]] | None = None,
    ignore_unknowns: bool = False,
    asuser: bool = False,
) -> None:
    from ..templating import get_templates
    from ..utils import maybe_closing, rmfiles

    templates = get_templates(template or "supervisor.ini")
    application_dir = application_dir or "."

    with maybe_closing(
        open(output, "w", encoding="utf-8") if isinstance(output, str) else output,
    ) as fp:
        try:
            for tplt in templates:
                supervisor(
                    tplt,
                    application_dir,
                    args,
                    check=check,
                    output=fp,
                    ignore_unknowns=ignore_unknowns,
                    help_args=help_args,
                    extra_params=extra_params,
                    checks=checks,
                    asuser=asuser,
                )
        except Exception as ex:
            if isinstance(output, str):
                rmfiles([output])
            raise ex


@config.command(name="supervisord", help=SUPERVISORD_HELP)  # noqa: C901
@config_options
@template_option
@click.option(
    "-d",
    "--app-dir",
    "application_dir",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    help="""location of repo or current directory""",
)
@click.argument("params", nargs=-1, required=False)
def supervisord_cmd(
    application_dir: str | None,
    params: list[str],
    template: str | None,
    no_check: bool,
    output: str | None,
) -> None:
    supervisord(
        template,
        application_dir or ".",
        params,
        check=not no_check,
        output=output,
        ignore_unknowns=True,
    )


@config.command(name="systemd-celery", help=CELERY_SYSTEMD_HELP)  # noqa: C901
@template_option
@asuser_option
@config_options
@click.option(
    "-d",
    "--app-dir",
    "application_dir",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    help="""location of repo or current directory""",
)
@click.argument("params", nargs=-1, required=False)
def systemd_celery_cmd(
    application_dir: str | None,
    params: list[str],
    template: str | None,
    no_check: bool,
    output: str | None,
    asuser: bool,
) -> None:
    import os
    from os.path import isfile

    from .utils import check_app_dir, check_venv_dir
    from .systemd import systemd

    application_dir = application_dir or "."

    def find_celery(params: dict[str, Any]) -> str | None:
        assert application_dir is not None
        for d in os.listdir(application_dir):
            fd = join(application_dir, d)
            if isdir(fd):
                for mod in ["celery", "tasks"]:
                    if isfile(join(fd, f"{mod}.py")):
                        return f"{d}.{mod}"
        return None

    def check_celery(venv: str) -> str | None:
        c = join(venv, "bin", "celery")
        if not os.access(c, os.X_OK | os.R_OK):
            return "please install celery!"
        return None

    systemd(
        template or "celery.service",
        application_dir or ".",
        params,
        help_args=SUPERVISORD_ARGS,
        check=not no_check,
        output=output,
        asuser=asuser,
        default_values=[("celery", find_celery)],
        checks=[
            ("application_dir", lambda _, v: check_app_dir(v)),
            ("venv", lambda _, v: check_venv_dir(v) or check_celery(v)),
        ],
    )



from __future__ import annotations

import subprocess

import click

from .cli import cli
from .utils import which


def restart_userd() -> list[tuple[str, int]]:
    """Restart any user systemd files"""
    import os
    from os.path import isdir, join

    from .utils import userdir as u

    userdir = u()

    status: list[tuple[str, int]] = []

    systemctl = which("systemctl")

    for f in os.listdir(userdir):
        if isdir(join(userdir, f)):  # skip directories
            continue
        if "@" in f:
            continue
        r = subprocess.run(
            [systemctl, "--user", "status", f],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=False,
        )
        # 4 unknown, 3 dead?
        if r.returncode == 3:
            # rep = r.stdout.strip()
            r = subprocess.run(
                [systemctl, "--user", "start", f],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=False,
            )
            status.append((f, r.returncode))
        elif r.returncode != 0:
            status.append((f, r.returncode))

    return status


@cli.command()
def systemd_restart() -> None:
    """Restart any dead *user* systemd services"""
    restarted = restart_userd()
    col = {0: "green", 2: "yellow", 1: "yellow"}
    for service, code in restarted:
        s = click.style(service, bold=True, fg=col.get(code, "red"))
        click.echo(f"restart[{code}]: {s}")
    if any(ok != 0 for _, ok in restarted):
        raise click.Abort()



from __future__ import annotations

import re
from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any
from urllib.parse import parse_qsl
from urllib.parse import unquote


@dataclass
class URL:
    drivername: str
    username: str | None = None
    password: str | None = None
    host: str | None = None
    port: int | None = None
    database: str | None = None
    query: dict[str, str | list[str] | None] | None = None


# direct ripoff of sqlalchemy.engine.url:make_url
# https://github.com/sqlalchemy/sqlalchemy/blob/7fdaf711dc6bd578f7becf45526dce70f523890d/lib/sqlalchemy/engine/url.py#L821


def make_url(name_or_url: str | URL) -> URL | None:
    if isinstance(name_or_url, URL):
        return name_or_url
    pattern = re.compile(
        r"""
            (?P<name>[\w\+]+)://
            (?:
                (?P<username>[^:/]*)
                (?::(?P<password>[^@]*))?
            @)?
            (?:
                (?:
                    \[(?P<ipv6host>[^/\?]+)\] |
                    (?P<ipv4host>[^/:\?]+)
                )?
                (?::(?P<port>[^/\?]*))?
            )?
            (?:/(?P<database>[^\?]*))?
            (?:\?(?P<query>.*))?
            """,
        re.X,
    )

    m = pattern.match(name_or_url)
    if m is None:
        return None

    components: dict[str, Any] = m.groupdict()
    if components["query"] is not None:
        query: dict[str, str | list[str]] = {}

        for key, value in parse_qsl(components["query"]):
            if key in query:
                query[key] = v = to_list(query[key])
                v.append(value)
            else:
                query[key] = value
    else:
        query = None  # type: ignore
    components["query"] = query

    if components["username"] is not None:
        components["username"] = unquote(components["username"])

    if components["password"] is not None:
        components["password"] = unquote(components["password"])

    ipv4host = components.pop("ipv4host")
    ipv6host = components.pop("ipv6host")
    components["host"] = ipv4host or ipv6host
    name = components.pop("name")

    if components["port"]:
        components["port"] = int(components["port"])

    return URL(drivername=name, **components)


def to_list(x: str | list[str]) -> list[str]:
    if not isinstance(x, Iterable) or isinstance(x, (str, bytes)):
        return [str(x)]
    elif isinstance(x, list):
        return x
    else:
        return list(x)



from __future__ import annotations

import subprocess

import click

from .cli import cli
from .utils import which


def rsync(src: str, tgt: str, verbose: bool = False) -> None:
    rsync_cmd = which("rsync")

    v = ["-v"] if verbose else []

    if not src.endswith("/"):
        src += "/"
    if tgt.endswith("/"):
        tgt = tgt[:-1]

    cmd = [rsync_cmd, "-a"] + v + ["--delete", src, tgt]
    subprocess.run(cmd, check=True)


@cli.command(name="rsync")
@click.option("-v", "--verbose", is_flag=True)
@click.argument("src")
@click.argument("tgt")
def rsync_cmd(src: str, tgt: str, verbose: bool) -> None:
    """Sync two directories on two possibly different machines.

    e.g.: footprint rsync my/folder chloe:/var/www/folder
    """
    rsync(src, tgt, verbose)



from __future__ import annotations

import os
import re
import subprocess
import sys
from collections.abc import Sequence
from os.path import isdir
from os.path import isfile
from os.path import join
from pathlib import Path
from typing import Any
from typing import Callable
from typing import TextIO
from typing import TypeVar

import click

from ..core import get_dot_env
from ..core import StaticFolder

F = TypeVar("F", bound=Callable[..., Any])

NUM = re.compile(r"^[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?$")

CONVERTER = Callable[[dict[str, Any]], Any]

CHECKTYPE = Callable[[str, Any], str | None]


class ArgError(Exception):
    def __init__(self, message: str):
        super().__init__()
        self.message = message


def fix_kv(
    key: str,
    values: list[str],
    convert: dict[str, CONVERTER] | None = None,
) -> tuple[str, Any]:
    # if key in {"gevent"}:  # boolean flag
    #     return ("gevent", True)
    if "" in values:
        raise ArgError(f"no value for {key}")
    key = key.replace("-", "_")
    if not values:  # simple key is True
        return (key, True)
    value = "=".join(values)

    def get_value(value: str) -> tuple[str, Any]:
        if key in {"user"}:  # user is a string!
            return (key, value)
        if value.isdigit():
            return (key, int(value))
        if value == "true":
            return (key, True)
        if value == "false":
            return (key, False)
        if NUM.match(value):
            return (key, float(value))
        return (key, value)

    key, v = get_value(value)
    if convert and key in convert:
        v = convert[key](v)
    return key, v


def fix_params(
    params: list[str],
    convert: dict[str, CONVERTER] | None = None,
) -> dict[str, Any]:
    from jinja2 import UndefinedError, Undefined

    def f(p: str) -> tuple[str, Any]:
        k, *values = p.split("=")
        if values == [""]:  # just skip 'key=' mistakes
            return k, Undefined
        return fix_kv(k, values, convert)

    try:
        return dict(f(p) for p in params)
    except ArgError as e:
        raise UndefinedError(e.message) from e


# KW = re.compile(r"^([\w_-]+)\s*:", re.M)


def get_known(help_args: dict[str, str]) -> set[str]:
    return {s.replace("-", "_") for s in help_args}


def url_match(directory: str, exclude: Sequence[str] | None = None) -> str:
    # scan directory and add any extra files directories
    # that are needed for location ~ /^(match1|match2|...) { .... }

    from ..config import get_config

    Config = get_config()

    if exclude is not None:
        sexclude = set(Config.exclude) | set(exclude)
    else:
        sexclude = set(Config.exclude)

    dirs = set(Config.static_dir)
    files = set(Config.static_files)
    for f in os.listdir(directory):
        if f in sexclude:
            continue
        tl = dirs if isdir(join(directory, f)) else files
        tl.add(f)

    d = "|".join(f.replace(".", r"\.") for f in dirs)
    f = "|".join(f.replace(".", r"\.") for f in files)
    return f"(^/({d})/|^({f})$)"


def find_favicon(application_dir: str) -> str | None:
    """Find directory with favicon.ico or robot.txt or other toplevel files"""
    from ..config import get_config

    Config = get_config()

    static = set(Config.static_files)
    for d, dirs, files in os.walk(application_dir, topdown=True):
        dirs[:] = [f for f in dirs if not f.startswith((".", "_"))]
        if d.startswith((".", "_")):
            continue
        for f in files:
            if f in static:
                return d
    return None


def check_app_dir(application_dir: str) -> str | None:
    if not isdir(application_dir):
        return f"not a directory: {application_dir}"
    return None


def check_venv_dir(venv_dir: str) -> str | None:
    if not isdir(venv_dir):
        return "venv: not a directory: {venv_dir}"

    py = join(venv_dir, "bin", "python")
    if not os.access(py, os.X_OK | os.R_OK):
        return f"venv: {venv_dir} does not have python installed!"
    return None


def footprint_config(application_dir: str) -> dict[str, Any]:
    def dot_env(f: str) -> dict[str, Any]:
        cfg = get_dot_env(f)
        if cfg is None:
            return {}
        return dict(
            fix_kv(k.lower(), [v])
            for k, v in cfg.items()
            if k.isupper() and v is not None
        )

    f = join(application_dir, ".flaskenv")
    if not isfile(f):
        return {}
    return dot_env(f)


def get_default_venv() -> str:
    venv = Path(sys.executable).parent.parent
    return str(venv)


def has_error_page(static_folders: list[StaticFolder]) -> StaticFolder | None:
    for s in static_folders:
        if "404.html" in os.listdir(s.folder):
            return s
    return None


def fixname(n: str) -> str:
    # return n.replace("\\", "\\\\")
    return n


def getgroup(username: str) -> str | None:
    username = username.replace("\\\\", "\\")
    try:
        # username might not exist on this machine
        ret = subprocess.check_output(["id", "-gn", username], text=True).strip()
        return fixname(ret)
    except subprocess.CalledProcessError:
        return None


def getuser() -> str | None:
    try:
        # username might not exist on this machine
        ret = subprocess.check_output(["id", "-un"], text=True).strip()
        return fixname(ret)
    except subprocess.CalledProcessError:
        return None


def make_args(argsd: dict[str, str], **kwargs: Any) -> str:
    from itertools import chain

    from ..config import get_config

    Config = get_config()

    def color(s: str) -> str:
        if Config.arg_color == "none":
            return s
        return click.style(s, fg=Config.arg_color)

    args = list((k, v) for k, v in chain(argsd.items(), kwargs.items()))

    argl = [(color(k), v) for k, v in args]
    aw = len(max(argl, key=lambda t: len(t[0]))[0]) + 1
    bw = len(max(args, key=lambda t: len(t[0]))[0]) + 1
    sep = "\n  " + (" " * bw)

    def fixd(d: str) -> str:
        dl = d.split("\n")
        return sep.join(dl)

    return "\n".join(f"{arg:<{aw}}: {fixd(desc)}" for arg, desc in argl)


def to_check_func(
    key: str,
    func: Callable[[Any], bool],
    msg: str,
) -> tuple[str, CHECKTYPE]:
    def f(k: str, val: Any) -> str | None:
        if func(val):
            return None
        return msg.format(**{key: val})

    return (key, f)


def to_output(res: str, output: str | TextIO | None = None) -> None:
    if not res.endswith("\n"):
        res += "\n"
    if output:
        if isinstance(output, str):
            with open(output, "w", encoding="utf-8") as fp:
                fp.write(res)

        else:
            output.write(res)
    else:
        click.echo(res)


def config_options(f: F) -> F:
    f = click.option(
        "-o",
        "--output",
        help="write to this file",
        type=click.Path(dir_okay=False),
    )(f)
    f = click.option("-n", "--no-check", is_flag=True, help="don't check parameters")(f)
    return f


# def su(f):
#     return click.option("--su", "use_su", is_flag=True, help="use su instead of sudo")(
#         f,
#     )


def asuser_option(f: F) -> F:
    f = click.option("-u", "--user", "asuser", is_flag=True, help="Install as user")(f)
    return f


def check_user(asuser: bool) -> None:
    if asuser:
        if os.geteuid() == 0:
            raise click.BadParameter(
                "can't install to user if running as root",
                param_hint="user",
            )


def template_option(f: F) -> F:
    return click.option(
        "-t",
        "--template",
        metavar="TEMPLATE_FILE",
        help="template file or directory of templates",
    )(f)


def asgi_option(f: F) -> F:
    return click.option(
        "--asgi",
        is_flag=True,
        help="run as asyncio (Quart|FastAPI)",
    )(f)



from __future__ import annotations

import os

import click

from .cli import cli


def img2ico(png: str, out: str) -> None:
    from PIL import Image  # type: ignore

    with open(png, "rb") as fp:
        im = Image.open(fp)

        im.thumbnail((128, 128), Image.ANTIALIAS)  # type: ignore

        size_tuples = [  # (256, 256),
            (128, 128),
            (64, 64),
            (48, 48),
            (32, 32),
            (24, 24),
            (16, 16),
        ]

        im.save(out, sizes=size_tuples)


@cli.command()
@click.option("-o", "--output", help="output filename")
@click.argument(
    "image",
    nargs=1,
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
)
def img_to_ico(image: str, output: str | None) -> None:
    "Convert a image file to an .ico file [**requires Pillow**]."
    from .utils import require_mod

    require_mod("PIL", "Pillow")

    # see https://anaconda.org/conda-forge/svg2png
    if output is None:
        out, _ = os.path.splitext(image)
        output = out + ".ico"
    img2ico(image, output)



from __future__ import annotations

import os
from os.path import dirname
from os.path import isabs
from os.path import join
from typing import Any
from typing import TYPE_CHECKING

import click

from .core import topath

if TYPE_CHECKING:
    from jinja2 import Template
    from jinja2 import UndefinedError
    from jinja2 import Environment


def templates_dir() -> str:
    return join(dirname(__file__), "templates")


def get_template_filename(name: str) -> str:
    return join(templates_dir(), name)


def get_env(application_dir: str | None = None) -> Environment:
    import datetime
    import sys

    from jinja2 import FileSystemLoader, StrictUndefined, UndefinedError, Environment

    def ujoin(*args: Any) -> str:
        for path in args:
            if isinstance(path, StrictUndefined):
                raise UndefinedError("undefined argument to join")
        return join(*[str(s) for s in args])

    def split(
        s: str | StrictUndefined,
        sep: str | None = None,
    ) -> list[str] | StrictUndefined:
        if isinstance(s, StrictUndefined):
            # raise UndefinedError("undefined argument to split")
            return s
        if sep is None:
            return s.split()
        return s.split(sep)

    def envf(envvar: str, default: str | None = None) -> str:
        if isinstance(envvar, StrictUndefined):
            raise UndefinedError("undefined argument to env")
        ret = os.environ.get(envvar, default)
        if ret is not None:
            return ret
        raise UndefinedError(f'unknown environment variable: "{envvar}"')

    def normpath(path: str | StrictUndefined) -> str | StrictUndefined:
        if isinstance(path, StrictUndefined):
            # raise UndefinedError("undefined argument to normpath")
            return path
        return topath(path)

    templates = [templates_dir()]
    if application_dir:
        templates = [application_dir, *templates]
    env = Environment(undefined=StrictUndefined, loader=FileSystemLoader(templates))

    def maybe_colon(s: str | StrictUndefined) -> str:
        if isinstance(s, StrictUndefined):
            return ""
        if not s:
            return s
        if s.endswith(":"):
            return s
        return s + ":"

    filt: dict[str, Any] = {
        "normpath": normpath,
        "split": split,
        "maybe_colon": maybe_colon,
        "env": envf,
    }

    glb: dict[str, Any] = {
        "join": ujoin,
        "cmd": " ".join(sys.argv),
        "now": lambda: datetime.datetime.now(datetime.timezone.utc),
    }

    env.filters.update(filt)  # type: ignore
    env.globals.update(glb)  # type: ignore

    return env


def get_template(
    template: str | Template,
    application_dir: str | None = None,
) -> Template:
    from jinja2 import Template

    if isinstance(template, Template):
        return template
    env = get_env(application_dir)
    if isabs(template):
        with open(template, encoding="utf8") as fp:
            t = env.from_string(fp.read())
            t.filename = template
            return t
    return env.get_template(template)


def get_templates(template: str) -> list[str | Template]:
    import os

    templates: list[str | Template]

    tm = topath(template)
    if os.path.isdir(tm):
        env = get_env(tm)
        templates = [env.get_template(f) for f in sorted(os.listdir(tm))]
    else:
        templates = [template]

    return templates


def undefined_error(
    exc: UndefinedError,
    template: Template,
    params: dict[str, Any],
) -> None:
    from .utils import get_variables

    msg = click.style(f"{exc.message}", fg="red", bold=True)
    names = sorted(params)
    variables = get_variables(template)
    missing = variables - set(names)
    if missing:
        s = "s" if len(missing) > 1 else ""
        mtext = click.style(
            f' variable{s} in template: {" ".join(missing)}',
            fg="yellow",
        )
    else:
        mtext = ""
    msg = click.style(f"{msg}:{mtext}")
    click.secho(msg, err=True)



from __future__ import annotations

from typing import Any
from typing import Iterator
from typing import TYPE_CHECKING

from .core import StaticFolder
from .utils import topath

if TYPE_CHECKING:
    from starlette.applications import Starlette


def is_starlette_app(app: Any) -> bool:
    try:
        from starlette.applications import Starlette  # type: ignore

        return isinstance(app, Starlette)
    except ImportError:
        return False


def get_starlette_static_folders(app: Starlette) -> Iterator[StaticFolder]:
    from typing import Sequence
    from starlette.staticfiles import StaticFiles
    from starlette.routing import Mount, Router, BaseRoute

    def findstatic(
        routes: Sequence[BaseRoute],
        prefix: str = "",
    ) -> Iterator[StaticFolder]:
        for r in routes:
            if isinstance(r, Mount):

                if isinstance(r.app, StaticFiles):
                    folder = r.app.directory
                    if not folder:
                        continue
                    folder = topath(str(folder))
                    path = prefix + r.path
                    rewrite = not folder.endswith(path)
                    yield StaticFolder(r.path, folder, rewrite)
                elif isinstance(r.app, Router):
                    yield from findstatic(r.app.routes, prefix + r.path)

    yield from findstatic(app.routes)



from __future__ import annotations

import click

from .cli import cli


@cli.command(
    epilog=click.style(
        """use e.g.: footprint secret >> instance/app.cfg""",
        fg="magenta",
    ),
)
@click.option("--size", default=32, help="size of secret in bytes", show_default=True)
@click.option("--str", "as_str", is_flag=True, help="output as string")
def secret(size: int, as_str: bool) -> None:
    """Generate secret keys for Flask apps"""
    from secrets import token_bytes

    def t() -> str | bytes:
        b = token_bytes(size)
        if not as_str:
            return b
        # return repr(b.hex())
        return repr(b.decode("latin1"))

    print("SECRET_KEY =", t())
    print("SECURITY_PASSWORD_SALT =", t())
